"""
数据类型类、枚举类、空类、类型别名等。
"""
import enum
from types import DynamicClassAttribute
from typing import Type, Union, Tuple, Dict, List

Throwable = Union[BaseException, Type[BaseException]]


class UNSET:
    pass


class OnionObject(object):

    def __translate_list(self, items: list) -> list:
        return list(
            self.__translate_list(list(item)) if isinstance(item, tuple)
            else self.__translate_list(item) if isinstance(item, list)
            else type(self)(item, self.__depth - 1) if isinstance(item, dict)
            else item
            for item in items
        )

    # OnionObject()
    def __new__(cls, dictionary: dict = None, depth: int = -1, **kvs):
        """
        将字典转化为对象，使得可以用点分法代替下标访问内容。

        支持还原为字典：

        >>> data: dict = ~OnionObject()

        支持直接更新字典：

        >>> obj: OnionObject = OnionObject()
        >>> obj: OnionObject = obj | dict()  # 等价于 obj |= dict()

        :param dictionary: 包含数据的字典。不符合标识符命名要求，
                           或者以双下划线 “__” 开头的键不会被收录。
        :param depth: 递归转化的层数。负数表示无限递归转化，正数表示递归层数，0无意义。
        :return: OnionObject 的对象。
        :raise TypeError: 属性名称必须是字符串。
        """
        if dictionary is None:
            dictionary = {}
        if not isinstance(dictionary, dict):
            raise TypeError  # pragma: no cover
        if depth == 0:
            return dictionary
        dictionary.update(kvs)

        self = object.__new__(cls)
        self.__depth = depth
        self.__ior__(dictionary)
        return self

    # OnionObject() | dict()
    def __or__(self, dictionary: dict) -> 'OnionObject':
        for k, v in dictionary.items():
            k = str(k)
            if not k.isidentifier() or k.startswith('__'):
                continue
            if isinstance(v, tuple):
                self.__setattr__(k, self.__translate_list(list(v[:])))
            elif isinstance(v, list):
                self.__setattr__(k, self.__translate_list(v[:]))
            elif isinstance(v, dict) and self.__depth:
                self.__setattr__(k, self.__class__(v, self.__depth - 1))
            else:
                self.__setattr__(k, v)
        return self

    # obj = OnionObject()
    # obj |= dict()
    __ior__ = __or__

    # ~OnionObject()
    def __invert__(self) -> dict:
        def obtain():
            prefix = f'_{type(self).__name__}__'
            for k, v in self.__dict__.items():
                if k.startswith('__') or k.startswith(prefix):
                    continue
                if isinstance(v, (list, tuple)):
                    yield k, [
                        ~i if isinstance(i, OnionObject) else i
                        for i in v
                    ]
                elif isinstance(v, OnionObject):
                    yield k, ~v
                else:
                    yield k, v

        return dict(obtain())

    # repr(OnionObject())
    def __repr__(self) -> str:
        prefix = f'_{type(self).__name__}__'
        attrs = ', '.join(
            f'{attr}={type(value).__name__}(...)'
            if isinstance(value, OnionObject) else
            f'{attr}={value!r}'
            for attr, value in self.__dict__.items()
            if not attr.startswith('__') or attr.startswith(prefix)
        )
        return f'OnionObject({attrs})'


# Little Endian
class RadixInteger(Tuple[int, ...]):

    def __new__(cls,
                x: Union[int, Tuple[int, ...], List[int], bytes, bytearray],
                n: int,
                be=False,
                negative=False) -> 'RadixInteger':
        """
        一个以元组表述各个数位的 N 进制整数。

        :param x: 一个int类型的整数，一个包含非负整数的元组，或者一个字节串。
        :param n: 进位制（的基数）。必须是一个大于等于 2 的正整数。
        :param be: 表示给定元组是否使用大端字节序。
        :param negative: 表示给定元组是否应当表示一个负数，或给定字节串是否使用二进制补码表示整数。
        """

        def dec2seq(i: int):
            while i >= n:
                yield i % n
                i //= n
            yield i

        if n < 2:
            raise ValueError('无法表述基数比 2 更低的进位制。')

        if isinstance(x, int):
            self = tuple.__new__(cls, dec2seq(abs(x)))
            self._radix = n
            self._integer = x

        elif isinstance(x, cls):
            self = tuple.__new__(cls, dec2seq(abs(x.numeric)))
            self._radix = n
            self._integer = x.numeric

        elif isinstance(x, (bytes, bytearray)):
            self = tuple.__new__(cls, x[::-1] if be else x)
            self._radix = 256
            self._integer = (
                int.from_bytes(x, 'big', signed=negative)
                if be else
                int.from_bytes(x, 'little', signed=negative)
            )

        elif isinstance(x, (tuple, list)):
            if min(x) < 0:
                raise ValueError('x 只能包含非负整数。')
            radix = max(x)
            if not radix <= n:
                raise ValueError(
                    f'x 的进位制最低是 {radix}，高于给定的 {n} 。'
                )
            pairs = zip(x, reversed(range(len(x))))
            self = tuple.__new__(cls, x[::-1] if be else x)
            self._radix = n
            self._integer = sum(map(lambda pair: pair[0] * n ** pair[1], pairs))
            self._integer = -self._integer if negative else self._integer

        else:
            raise TypeError(
                f'{cls.__name__}() 不接受 {type(x).__name__} 类型的参数。'
            )

        return self

    @property
    def radix(self):
        """整数的进位制。"""
        return self._radix

    @property
    def numeric(self):
        """对应的以阿拉伯数字表述的十进制整数。"""
        return self._integer

    def map2str(self, mapping: Union[str, Dict[int, str]], be=True) -> str:
        """
        按照规则将每一位数映射到一个字符串中。
        """
        return ''.join(map(lambda i: mapping[i], self[::-1] if be else self))

    def map2bytes(self, mapping: Union[bytes, Dict[int, bytes]], be=True) -> bytes:
        """
        按照规则将每一位数映射到一个字节串中。
        """
        return bytes(map(lambda i: mapping[i], self[::-1] if be else self))


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class ChoicesMeta(enum.EnumMeta):
    """用于创建带有标签文本的枚举的类（元类）。"""

    def __new__(metacls, classname, bases, classdict, **kwds):
        labels = []
        for key in classdict._member_names:
            value = classdict[key]
            if (
                    isinstance(value, (list, tuple))
                    and len(value) > 1
                    and isinstance(value[-1], (str,))
            ):
                *value, label = value
                value = tuple(value)
            else:
                label = key.replace("_", " ").title()
            labels.append(label)
            # Use dict.__setitem__() to suppress defenses against double
            # assignment in enum's classdict.
            dict.__setitem__(classdict, key, value)
        cls = super().__new__(metacls, classname, bases, classdict, **kwds)
        for member, label in zip(cls.__members__.values(), labels):
            member._label_ = label
        return enum.unique(cls)

    def __contains__(cls, member):
        if not isinstance(member, enum.Enum):
            # Allow non-enums to match against member values.
            return any(x.value == member for x in cls)
        return super().__contains__(member)

    @property
    def names(cls):
        empty = ["__empty__"] if hasattr(cls, "__empty__") else []
        return empty + [member.name for member in cls]

    @property
    def choices(cls):
        empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
        return empty + [(member.value, member.label) for member in cls]

    @property
    def labels(cls):
        return [label for _, label in cls.choices]

    @property
    def values(cls):
        return [value for value, _ in cls.choices]


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class Choices(enum.Enum, metaclass=ChoicesMeta):
    """用于创建带有标签文本的枚举。"""

    @DynamicClassAttribute
    def label(self):
        return self._label_

    @property
    def do_not_call_in_templates(self):
        return True

    def __str__(self):
        """
        Use value when cast to str, so that Choices set as model instance
        attributes are rendered as expected in templates and similar contexts.
        """
        return str(self.value)

    # A similar format was proposed for Python 3.10.
    def __repr__(self):
        return f"{self.__class__.__qualname__}.{self._name_}"


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class IntegerChoices(int, Choices):
    """用于创建值是整数的带有标签文本的枚举。"""

    pass


# copy from https://github.com/django/django/blob/stable/4.2.x/django/db/models/enums.py
class TextChoices(str, Choices):
    """用于创建值是字符串的带有标签文本的枚举。"""

    def _generate_next_value_(name, start, count, last_values):
        return name


class ItemsMeta(enum.EnumMeta):
    """用于创建带有任意属性的枚举的类（元类）。"""

    def __new__(metacls, classname, bases, classdict, **kwds):
        # 获取属性名（pks）
        if '__properties__' not in classdict:
            raise AttributeError(
                f'{classname} 使用了 {metacls.__name__}，'
                f'因此必须定义一个名为 __properties__ 的属性。'
            )
        pks = classdict['__properties__']
        if not isinstance(pks, (tuple, list)):
            raise TypeError(
                f'{classname}.__properties__ 的值只允许是 tuple 或 list 。'
            )
        if not all(isinstance(pk, str) and not pk.startswith('_') for pk in pks):
            raise TypeError(
                f'{classname}.__properties__ 的值必须是字符串且不以下划线 “_” 开头。'
            )
        pks = tuple(f'_{pk}_' for pk in pks)
        qty = len(pks) + 1  # 等号右侧所有元素的总数

        # 获取属性值（pvs）和枚举值（value）
        pvs_list = []
        for key in classdict._member_names:
            value = classdict[key]
            pvs = ()
            if isinstance(value, tuple) and len(value) == qty:
                value, *pvs = value
                pvs = tuple(pvs)
            pvs_list.append(pvs)
            dict.__setitem__(classdict, key, value)

        # 为每个枚举添加私有属性
        cls = super().__new__(metacls, classname, bases, classdict, **kwds)
        for member, pvs in zip(cls.__members__.values(), pvs_list):
            member.__dict__.update(zip(pks, pvs))

        return enum.unique(cls)

    def __contains__(cls, member):
        if not isinstance(member, enum.Enum):
            # Allow non-enums to match against member values.
            return any(x.value == member for x in cls)
        return super().__contains__(member)

    def __getattr__(cls, name):
        if not isinstance(name, str):
            raise TypeError
        if name[:-1] in cls.__properties__ and name.endswith('s'):
            return [getattr(member, name[:-1]) for member in cls]
        if name[:-2] in cls.__properties__ and name.endswith('es'):
            return [getattr(member, name[:-2]) for member in cls]
        return object.__getattribute__(cls, name)

    @property
    def names(cls):
        empty = ["__empty__"] if hasattr(cls, "__empty__") else []
        return empty + [member.name for member in cls]

    @property
    def values(cls):
        empty = [None] if hasattr(cls, "__empty__") else []
        return empty + [member.value for member in cls]

    @property
    def choices(cls):
        if 'label' not in cls.__properties__:
            raise AttributeError(
                '如需使用 .choices 属性，必须在 __properties__ 中'
                '添加一个名为 "label" 的属性，且必须保证枚举值中有相应的属性值。'
            )
        empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
        return empty + [(member.value, member.label) for member in cls]


class Items(enum.Enum, metaclass=ItemsMeta):
    """
    用于创建带有任意属性的枚举。

    >>> from enum import Enum
    >>>
    >>> class YearInSchool(Items):
    >>>     FRESHMAN = 1, 'FR', 0xE35314, 'Freshman'
    >>>     SOPHOMORE = 2, 'SO', 0xED15B4, 'Sophomore'
    >>>     JUNIOR = 3, 'JR', 0x9B3CED, 'Junior'
    >>>     SENIOR = 4, 'SR', 0xA0408E, 'Senior'
    >>>     GRADUATE = 5, 'GR', 0x518CCC, 'Graduate'
    >>>
    >>>     __properties__ = 'code', 'color', 'label'
    >>>
    >>>     @property
    >>>     def code(self) -> str:
    >>>         return self._code_
    >>>
    >>>     @property
    >>>     def color(self) -> int:
    >>>         return self._color_
    >>>
    >>>     @property
    >>>     def label(self) -> str:
    >>>         return self._label_
    >>>
    >>> print(YearInSchool.SENIOR.name)  # 'SENIOR'
    >>> print(YearInSchool.SENIOR.value)  # 4
    >>> print(YearInSchool.SENIOR.code)  # 'SR'
    >>> print(hex(YearInSchool.SENIOR.color))  # '0xa0408e'
    >>> print(YearInSchool.SENIOR.label)  # 'Senior'
    >>>
    >>> print(YearInSchool.names)  # ['FRESHMAN', 'SOPHOMORE', ...]
    >>> print(YearInSchool.values)  # [1, 2, 3, 4, 5]
    >>> print(YearInSchool.codes)  # ['FR', 'SO', 'JR', 'SR', 'GR']
    >>> print(YearInSchool.colors)  # [14897940, 15537588, ...]
    >>> print(YearInSchool.labels)  # ['Freshman', 'Sophomore', ...]
    """
    __properties__ = ()

    def __repr__(self):
        # 枚举也算一种常量，直接按路径查找即可，故舍去附加的属性
        return f"{self.__class__.__qualname__}.{self._name_}"



"""
纯工具类。
"""
import sys
from datetime import datetime
from functools import wraps
from io import TextIOWrapper
from logging import Logger, DEBUG, getLevelName
from typing import Tuple, Optional, Union, NoReturn, TextIO

from .converters import delta2s, represent


class BearTimer(object):
    fmt = '[{head:%H:%M:%S.%f}] [{level}] [{title}] [{total:.6f} +{delta:.6f}]: {msg}'
    level = DEBUG

    def __init__(self, title: str = None, output: Union[TextIOWrapper, TextIO, Logger] = None):
        """
        熊牌计时器。对代码运行进行计时，并打印时间和提示。

        比较简单的用法是使用 with 语句包裹需要计时的部分，
        它会在开始执行前启动计时，在执行完毕或出现异常而离开with后停止计时。

        >>> with BearTimer() as bear:
        >>>     summary = 0
        >>>     for i in range(1000):
        >>>         if not i % 67:
        >>>             bear.step(f'loop to {i} now.')
        >>>         summary += i

        如果需要对一整个函数的运行进行计时，那么可以将计时器作为装饰器使用：

        >>> @BearTimer()
        >>> def prepare_order(request, *args, **kwargs):
        >>>     # 业务逻辑
        >>>     pass

        如果不方便使用 with 包裹，可以直接实例化一个类对象。
        每一个对象都是独立的计时器，互不影响。

        >>> bear = BearTimer()
        >>> bear.start()
        >>> bear.stop()

        :param title: 计时器的标题，用以标明输出信息归属于哪个计时器。默认从打印消息时的上下文中获取。
        :param output: 消息打印到哪里去。默认是系统标准输出。可以是一个文件IO对象，也可以是一个日志记录器。
        """
        self._start = None  # 计时开始时间
        self._point = None  # 中途标记时间（用于计算距离上次标记过去了多久）
        self._title = title
        self.output = output
        if not self._title:
            if hasattr(sys, '_getframe'):
                self._title = sys._getframe().f_back.f_code.co_name
            else:
                self._title = '<UNTITLED>'

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self._title = func.__name__
            self.start()
            returns = func(*args, **kwargs)
            self.stop()
            return returns

        return wrapper

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.stop()

    def handle(self, full_message: str) -> NoReturn:
        """
        将消息发送到指定位置。这个位置允许是文件IO对象，或者是日志记录器。

        不应直接使用此方法。

        :param full_message: 完整的消息内容。
        :return: 无。
        """
        if hasattr(self.output, 'log') and callable(self.output.log):
            self.output: Logger
            self.output.log(self.level, full_message)

        elif hasattr(self.output, 'write') and callable(self.output.write):
            self.output: TextIOWrapper
            print(full_message, file=self.output)

        else:
            print(full_message)

    def record(self, msg: str = '', *args, **kwargs) -> datetime:
        """
        打印一行消息，标明此刻的时间、经历的时长、距上次打印的时长、计时器标题，以及附加的消息。

        不应直接使用这个方法。

        :param msg: 要附加的消息。默认为空文本。
        :param args: 其它需要打印的位置参数。
        :param kwargs: 其它需要打印的关键字参数。
        :return: 此时的时刻。
        """
        now = datetime.now()
        total = delta2s(now - self._start) if self._start else 0
        delta = delta2s(now - self._point) if self._point else 0
        title = self._title if self._title else ''
        message = self.fmt.format(
            head=now, total=total, delta=delta, title=title, msg=msg,
            level=getLevelName(self.level), *args, **kwargs,
        )
        self._point = now
        self.handle(message)
        return now

    def start(self, msg='Starting...'):
        """
        开始计时。

        如果计时尚未结束，会重新开始计时，中途记录的标记也会被清除。

        :param msg: 要附加的消息。
        :return: 计时器对象自身。
        """
        self._point = None
        self._start = datetime.now()
        self.record(msg)
        return self

    def step(self, msg='') -> Tuple[Optional[datetime], datetime]:
        """
        中途标记。

        用此方法可以计算出距离上一次标记/开始计时过去了多久。

        如果计时器尚未开始计时，那么过去的时间始终是 0 秒。

        :param msg: 要附加的消息。
        :return: 二元元组。包含上一次标记/开始计时的时刻，和此时的时刻。
        """
        prev = self._point
        curr = self.record(msg)
        return prev, curr

    def stop(self, msg='Stopped.') -> datetime:
        """
        结束计时。

        :param msg: 要附加的消息。
        :return: 此时的时刻。
        """
        curr = self.record(msg)
        self._start = None
        return curr


class ReprMixin(object):
    """
    生成通用representation的工具类。

    格式类似于 ``<User(1) female name="meow" age=12 birth=[2012-01-23]>``，
    包含自身类名、主键、标签、属性名和值。

    直接或间接继承时请放在第一父类。
    """

    def __repr__(self) -> str:
        kls = self._obtain_kls()
        pkv = self._obtain_pk()
        tags = self._obtain_tags()
        attrs = self._obtain_attrs()
        content = (
                (f'({pkv})' if pkv else '') +
                (f' {tags}' if tags else '') +
                (f' {attrs}' if attrs else '')
        )
        return f'<{kls}{content}>'

    class AttributeMeta:
        """
        用于控制生成 representation 时需要带上哪些属性。

        注意：这个内部类不会被实例化！

        AttributeMeta 的变量代表你的类对象在运行时已经存在的属性，
        变量值应当是一个字符串，表示生成 representation 时这个属性的名称是什么。

        AttributeMeta 的变量允许接收一个返回值为字符串的函数作为类型注解，
        用于转换你的类对象的属性值，并直接作为 representation 里这个属性的值。
        """

    def _obtain_attrs(self) -> str:
        def obtain():
            for attr, name in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, represent)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                yield f'{name}={value}'

        attributes = self.AttributeMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ' '.join(filter(None, obtain()))

    class TagMeta:
        """
        用于控制生成 representation 时需要带上哪些标签。

        注意：这个内部类不会被实例化！

        TagMeta 的变量代表你的类对象在运行时已经存在的属性，变量值可以是
          - 一个字符串，表示这个属性为 ``True`` 时 representation 里才会出现的标签的名称；
            如果属性为 ``False`` 则不会出现这个标签。
          - 一个元组且仅有两个字符串，表示这个属性分别为
            ``False`` 和 ``True`` 时 representation 里会出现的标签的名称。
          - 一个列表或一个字典，则使用属性值对这个列表或字典进行取值，
            以此作为 representation 里出现的标签的名称。

        TagMeta 的变量允许接收一个返回值为字符串的函数作为类型注解，
        用于进一步转换你的类对象的属性值，若未提供，默认使用 bool() 来转换。
        """

    def _obtain_tags(self) -> str:
        def obtain():
            for attr, option in attributes.items():
                if attr.startswith('_'):
                    continue
                mapper = annotations.get(attr, None)
                value = getattr(self, attr)
                value = mapper(value) if callable(mapper) else value
                if isinstance(option, str) and value:
                    yield option
                elif isinstance(option, tuple):
                    yield option[bool(value)]
                elif isinstance(option, (list, dict)):
                    yield option[value]
                else:
                    pass  # pragma: no cover

        attributes = self.TagMeta.__dict__
        annotations = attributes.get('__annotations__', {})
        return ' '.join(filter(None, obtain()))

    def _obtain_pk(self) -> str:
        if hasattr(self, 'pk'):
            pk = self.pk
        elif hasattr(self, 'id'):
            pk = self.id
        else:
            return ''
        return pk if isinstance(pk, str) else represent(pk)

    def _obtain_kls(self) -> str:
        return self.__class__.__name__



"""
用于将一种值转换为另一种值的转换器。
"""
import re
from datetime import timedelta, datetime, date
from decimal import Decimal
from typing import Callable, Any, Tuple, Union
from uuid import UUID

from .typeclasses import Throwable, UNSET


def remove_exponent(d: Decimal):
    """
    去除十进制小数（Decimal）的尾导零。

    非原创代码，出自：
    https://docs.python.org/zh-cn/3/library/decimal.html#decimal-faq
    """
    return d.quantize(Decimal(1)) if d == d.to_integral() else d.normalize()


def delta2hms(delta: timedelta) -> Tuple[int, int, float]:
    """
    将时间增量转换为时分秒格式，其中秒钟以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 一个三元元组。
    """
    h = delta.seconds // 3600
    m = delta.seconds % 3600 // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return h, m, s


def delta2ms(delta: timedelta) -> Tuple[int, float]:
    """
    将时间增量转换为分秒格式，其中秒钟以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 二元元组。前者用一个整数表示分钟数，
             后者用一个小数表示秒钟数和纳秒数。
    """
    m = delta.seconds // 60
    s = delta.seconds % 60 + delta.microseconds / 1000000
    return m, s


def delta2s(delta: timedelta) -> float:
    """
    将时间增量转换为秒钟数，以小数形式包含毫秒和微秒。

    :param delta: 时间增量。
    :return: 一个小数。
    """
    return delta.seconds + delta.microseconds / 1000000


def represent(value) -> str:
    """
    将任意值转换为一个易于阅读的字符串。

    也是 ReprMixin 的默认格式化函数。

    默认使用 repr() 函数进行转换。如果自定义的类需要实现被此函数转换，请重写 .__repr__() 方法。

    :param value: 任意值。
    :return: 字符串。
    """
    if isinstance(value, str):
        return f'"{value}"'
    elif isinstance(value, timedelta):
        return f'[{value.days}d+{value.seconds}.{value.microseconds:06d}s]'
    elif isinstance(value, datetime):
        return f'[{value:%Y-%m-%d %H:%M:%S,%f}]'
    elif isinstance(value, date):
        return f'[{value:%Y-%m-%d}]'
    elif isinstance(value, UUID):
        return value.hex
    else:
        return repr(value)


def datasize(literal: str) -> Union[int, float]:
    """
    将一个字面量转换为字节数目。

    支持的单位包括：
      - B、b
      - KB、KiB、Kb、Kib
      - MB、MiB、Mb、Mib
      - GB、GiB、Gb、Gib
      - TB、TiB、Tb、Tib
      - 以此类推……

    - 1 B == 8 b
    - 1 MB == 1000 KB
    - 1 MiB == 1024 KiB

    :param literal: 一个整数后缀数据大小的单位。
    :return:
    """
    if not isinstance(literal, str):
        raise TypeError(
            '不支持解析一个非字符串类型的值。'  # pragma: no cover
        )

    pattern = re.compile(r'^([0-9]+)\s*([KMGTPEZY]?)(i?[Bb])$')
    result = re.fullmatch(pattern, literal)

    if result is None:
        return 0

    base = int(result.group(1))
    shift = 'BKMGTPEZY'.index(result.group(2))
    power = (1024 if 'i' in result.group(3) else 1000) ** shift
    power = (power / 8) if 'b' in result.group(3) else power

    return base * power


def true(value) -> bool:
    """
    将HTTP请求中 query 部分的参数值转换为 Python 的逻辑值。

    :param value: query 中的参数值。
    :return: True 或 False。
    """
    return value in ('true', 'True', 'TRUE', 1, True)


def safecast(mapper: Callable, raw, *errs: Throwable, default=None) -> Any:
    """
    转换一个值，转换失败时返回default，确保不会发生指定异常。

    默认捕获以下异常，可以通过errs参数追加更多异常：
        - TypeError
        - ValueError
        - KeyboardInterrupt

    :param mapper: 用来转换raw的转换器。如果转换器不可调用，将直接返回默认值。
    :param raw: 被转换的值。
    :param errs: 需要捕获的其它异常类或异常对象。应当提供可被 except 语句接受的值。
    :param default: 默认值。即使不提供也会默认返回 None 而不会抛出异常。
    :return: 转换后的值。如若捕获到特定异常将返回默认值。
    """
    exceptions = tuple(
        exc for exc in errs
        if exc.__class__ is type
        and issubclass(exc, BaseException)
        or isinstance(exc, BaseException)
    )
    if not callable(mapper):
        return default  # pragma: no cover
    try:
        return mapper(raw)
    except (TypeError, ValueError, KeyboardInterrupt) + exceptions:
        return default


class SafeCaster:

    def __init__(self, raw: Any = UNSET):
        """
        创建一个安全转换器，用于转换某个值而不发生特定异常。若不提供初值，则第一个转换器也必须允许无参调用。
        """
        self._value = raw
        self._default = None
        self._converters = tuple()
        self._exceptions = tuple()

    def __call__(self, raw: Any = UNSET) -> 'SafeCaster':
        self._value = raw
        return self

    def catch(self, *exceptions: Throwable) -> 'SafeCaster':
        """
        置入需要捕获的异常。应当提供可被 except 语句接受的值。
        """
        self._exceptions = tuple(
            exc for exc in exceptions
            if exc.__class__ is type
            and issubclass(exc, BaseException)
            or isinstance(exc, BaseException)
        )
        return self

    def by(self, *mappers: Callable) -> 'SafeCaster':
        """
        置入转换器。若未提供初值，则第一个转换器也必须允许无参调用。
        """
        self._converters = tuple(
            mapper for mapper in mappers
            if callable(mapper)
        )
        return self

    def get(self, default=None) -> Any:
        """
        执行转换并返回转换结果。若触发置入的异常或没有置入转换器则返回默认值；若触发未被置入的异常将原样抛出。
        """
        if len(self._converters) <= 0:
            return default

        if self._value is UNSET:
            result = self._converters[0]()
            converters = self._converters[1:]
        else:
            result = self._value
            converters = self._converters

        try:
            for converter in converters:
                result = converter(result) if callable(converter) else result
            return result
        except self._exceptions:
            return default


safecasts = SafeCaster().catch(TypeError, ValueError, KeyboardInterrupt)



"""
中国行政区划相关。
"""

from .typeclasses import IntegerChoices, Items


class Region(IntegerChoices):
    NORTH = 1, '华北'
    NORTHEAST = 2, '东北'
    EAST = 3, '华东'
    SOUTH_CENTRAL = 4, '中南'
    SOUTHWEST = 5, '西南'
    NORTHWEST = 6, '西北'
    HMT = 0, '港澳台'


class Province(Items):
    """
    中国省级行政区。
    """
    BEIJING = '11', 11, '京', 'BJ', '北京', Region.NORTH, '北京市'
    TIANJIN = '12', 12, '津', 'TJ', '天津', Region.NORTH, '天津市'
    HEBEI = '13', 13, '冀', 'HE', '河北', Region.NORTH, '河北省'
    SHANXI = '14', 14, '晋', 'SX', '山西', Region.NORTH, '山西省'
    NEI_MONGOLIA = '15', 15, '蒙', 'NM', '内蒙古', Region.NORTH, '内蒙古自治区'
    LIAONING = '21', 21, '辽', 'LN', '辽宁', Region.NORTHEAST, '辽宁省'
    JILIN = '22', 22, '吉', 'JL', '吉林', Region.NORTHEAST, '吉林省'
    HEILONGJIANG = '23', 23, '黑', 'HL', '黑龙江', Region.NORTHEAST, '黑龙江省'
    SHANGHAI = '31', 31, '沪', 'SH', '上海', Region.EAST, '上海市'
    JIANGSU = '32', 32, '苏', 'JS', '江苏', Region.EAST, '江苏省'
    ZHEJIANG = '33', 33, '浙', 'ZJ', '浙江', Region.EAST, '浙江省'
    ANHUI = '34', 34, '皖', 'AH', '安徽', Region.EAST, '安徽省'
    FUJIAN = '35', 35, '闽', 'FJ', '福建', Region.EAST, '福建省'
    JIANGXI = '36', 36, '赣', 'JX', '江西', Region.EAST, '江西省'
    SHANDONG = '37', 37, '鲁', 'SD', '山东', Region.EAST, '山东省'
    HENAN = '41', 41, '豫', 'HA', '河南', Region.SOUTH_CENTRAL, '河南省'
    HUBEI = '42', 42, '鄂', 'HB', '湖北', Region.SOUTH_CENTRAL, '湖北省'
    HUNAN = '43', 43, '湘', 'HN', '湖南', Region.SOUTH_CENTRAL, '湖南省'
    GUANGDONG = '44', 44, '粤', 'GD', '广东', Region.SOUTH_CENTRAL, '广东省'
    GUANGXI = '45', 45, '桂', 'GX', '广西', Region.SOUTH_CENTRAL, '广西壮族自治区'
    HAINAN = '46', 46, '琼', 'HI', '海南', Region.SOUTH_CENTRAL, '海南省'
    CHONGQING = '50', 50, '渝', 'CQ', '重庆', Region.SOUTHWEST, '重庆市'
    SICHUAN = '51', 51, '川', 'SC', '四川', Region.SOUTHWEST, '四川省'
    GUIZHOU = '52', 52, '黔', 'GN', '贵州', Region.SOUTHWEST, '贵州省'
    YUNNAN = '53', 53, '滇', 'YZ', '云南', Region.SOUTHWEST, '云南省'
    XIZANG = '54', 54, '藏', 'XY', '西藏', Region.SOUTHWEST, '西藏自治区'
    SHAANXI = '61', 61, '陕', 'SN', '陕西', Region.NORTHWEST, '陕西省'
    GANSU = '62', 62, '甘', 'GS', '甘肃', Region.NORTHWEST, '甘肃省'
    QINGHAI = '63', 63, '青', 'QH', '青海', Region.NORTHWEST, '青海省'
    NINGXIA = '64', 64, '宁', 'NX', '宁夏', Region.NORTHWEST, '宁夏回族自治区'
    XINJIANG = '65', 65, '新', 'XJ', '新疆', Region.NORTHWEST, '新疆维吾尔自治区'
    TAIWAN = '71', 71, '台', 'TW', '台湾', Region.HMT, '台湾省'
    HONG_KONG = '81', 81, '港', 'HK', '香港', Region.HMT, '香港特别行政区'
    MACAO = '82', 82, '澳', 'MO', '澳门', Region.HMT, '澳门特别行政区'

    __properties__ = 'numeric', 'short', 'code', 'nick', 'region', 'label'

    @property
    def numeric(self) -> int:
        """整数形式的区划代码。（2位）"""
        return self._numeric_

    @property
    def short(self) -> str:
        """简称。"""
        return self._short_

    @property
    def code(self) -> str:
        """字母码（见 GB/T 2260--2007）。"""
        return self._code_

    @property
    def nick(self) -> str:
        """常用别称。"""
        return self._nick_

    @property
    def label(self) -> str:
        """全称。"""
        return self._label_

    @property
    def region(self) -> Region:
        """所在大区。"""
        return self._region_

    @property
    def id_code(self) -> str:
        """行政区划代码。（6位）"""
        return self.value[0].ljust(6, '0')

    @property
    def statistics_code(self) -> str:
        """统计用区划代码。（12位）"""
        return self.value[0].ljust(12, '0')

    def __int__(self) -> int:
        return self._numeric__



from .checkers import start
from .converters import (
    remove_exponent,
    delta2hms,
    delta2ms,
    delta2s,
    represent,
    datasize,
    datasize as dsz,
    true,
    SafeCaster,
    safecast,
    safecasts,
)
from .divisions import (
    Region,
    Province,
)
from .generators import (
    randbytes,
    randb62,
    randb64,
    SnowflakeWorker,
    SnowflakeMultiWorker,
    SnowflakeSingleWorker,
)
from .typeclasses import (
    OnionObject,
    RadixInteger,
    ChoicesMeta,
    Choices,
    IntegerChoices,
    TextChoices,
    ItemsMeta,
    Items,
)
from .utils import (
    BearTimer,
    ReprMixin,
)

__version__ = (0, 2, 13)
version = '0.2.13'

# Django makemigrations 会在 CreateModel 里插入参数
# bases=(zeraora.utils.ReprMixin, models.Model)
# 导致对 zeraora 产生依赖，这里改变 ReprMixin 所在包地址，
# 尽量减少因为改变内部包结构导致对外部的影响。
ReprMixin.__module__ = 'zeraora'
